# $Id: //Depot/Source/SFW/riptable/Python/core/riptable/tests/test_multiset.py#4 $
import unittest

import numpy as np

# from io import StringIO
from riptable import Dataset, Multiset, Struct


class Multiset_Test(unittest.TestCase):
    @staticmethod
    def get_basic_dataset(keyfield_value=None, nr=10):
        ds = Dataset({_k: list(range(_i * nr, (_i + 1) * nr)) for _i, _k in enumerate("abcdefghijklmnop")})
        if keyfield_value is not None:
            ds.keyfield = keyfield_value
        return ds

    def test_ctor_01(self):
        dd = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("bac")}
        ms = Multiset(dd)
        self.assertEqual(list(ms.keys()), ["b", "a", "c"])
        for _i, _k in enumerate("bac"):
            self.assertEqual(ms[_k].keyfield[0], _i)
            self.assertEqual(getattr(ms, _k).keyfield[0], _i)
        with self.assertRaises(TypeError):
            _ = Multiset(ms)
        with self.assertRaises(TypeError):
            _ = Multiset(list(dd.items()))
        assert Multiset(None).shape == (
            0,
            0,
        ), f"Multiset default argument for 'None' input value should create an empty Multiset."
        ms = Multiset()
        self.assertEqual(ms.shape, (0, 0))
        ms = Multiset({})
        self.assertEqual(ms.shape, (0, 0))

    def test_ctor_02(self):
        dd = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("bac")}
        dd["d"] = self.get_basic_dataset(3)
        ms = Multiset(dd)
        self.assertEqual(list(ms.keys()), ["b", "a", "c", "d"])
        dd["d"] = Multiset({_k: self.get_basic_dataset(_i) for _i, _k in enumerate("bac")})
        ms = Multiset(dd)
        self.assertEqual(list(ms.keys()), ["b", "a", "c", "d"])
        # Number of rows in contained datasets must match
        dd["d"] = self.get_basic_dataset(3, nr=9)
        with self.assertRaises(ValueError):
            _ = Multiset(dd)
        # Number of rows in contained datasets and multisets must match
        dd["d"] = Multiset({})
        with self.assertRaises(ValueError):
            _ = Multiset(dd)
        dd["d"] = Multiset({_k: self.get_basic_dataset(_i, nr=9) for _i, _k in enumerate("bac")})
        with self.assertRaises(ValueError):
            _ = Multiset(dd)
        # Type of input must be valid
        dd["d"] = Struct({})
        with self.assertRaises(TypeError):
            _ = Multiset(dd)
        dd["d"] = None
        with self.assertRaises(TypeError):
            _ = Multiset(dd)
        dd["d"] = 1
        with self.assertRaises(TypeError):
            _ = Multiset(dd)

    def test_modify(self):
        dd = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("bac")}
        ms = Multiset(dd)
        self.assertEqual(list(ms.keys()), list("bac"))
        with self.assertRaises(ValueError):
            ms.d = Dataset({})
        ms.d = Dataset({"a": range(10)})
        self.assertEqual(list(ms.keys()), list("bacd"))
        with self.assertRaises(TypeError):
            ms.e = Struct({})
        with self.assertRaises(TypeError):
            ms.d = Struct({})
        self.assertEqual(list(ms.keys()), list("bacd"))
        self.assertIsInstance(ms.d, Dataset)
        ms._lock()
        with self.assertRaises(AttributeError):
            ms.e = Dataset({})
        ms._unlock()
        with self.assertRaises(ValueError):
            ms.e = Dataset({})
        ms.e = Dataset({"a": range(10)})
        self.assertEqual(list(ms.keys()), list("bacde"))
        self.assertEqual(ms.e.shape, (10, 1))
        self.assertIs(ms.e, ms[4])
        self.assertEqual(ms[4].shape, (10, 1))
        # Added Dataset must have same number of rows
        with self.assertRaises(ValueError):
            ms["f"] = Dataset({"a": 1, "b": 2})
        ms["f"] = Dataset({"a": range(10), "b": range(10)})
        ms[5] = Dataset({"a": [3] * 10, "b": [4] * 10})
        with self.assertRaises(IndexError):
            ms[6] = Dataset({"a": [3] * 10, "b": [4] * 10})
        self.assertEqual(list(ms.keys()), list("bacdef"))
        with self.assertRaises(IndexError):
            ms[["e", "g"]] = [Dataset({"a": 5, "b": 6}), Dataset({"a": 7, "b": 8})]
        ms[["e", "f"]] = [
            Dataset({"a": [5] * 10, "b": [6] * 10}),
            Dataset({"a": [7] * 10, "b": [8] * 10}),
        ]

    def test_multiget(self):
        # First testing case where all the contained Datasets are different:
        dd1 = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("ABC")}
        dd1["A"].col_remove(list("abc"))
        dd1["B"].col_remove(list("def"))
        dd1["D"] = Multiset({"SUB": dd1["C"]})
        ms1 = Multiset(dd1)
        ms2 = ms1.multiget(list("aeg"))
        self.assertEqual(list(ms1.keys()), list("ABCD"))
        self.assertEqual(list(ms2.keys()), list("ABC"))
        self.assertEqual(list(ms2.A.keys()), list("eg"))
        self.assertEqual(list(ms2.B.keys()), list("ag"))
        self.assertEqual(list(ms2.C.keys()), list("aeg"))
        ms3 = ms1.multiget(slice(2, 15, 4))
        self.assertEqual(list(ms3.keys()), list("ABC"))
        self.assertEqual(list(ms3.A.keys()), list("fjn"))
        self.assertEqual(list(ms3.B.keys()), list("cjn"))
        self.assertEqual(list(ms3.C.keys()), list("cgko"))
        # Now, when they all match:
        dd2 = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("ABCD")}
        ms4 = Multiset(dd2)
        self.assertEqual(list(ms4.keys()), list("ABCD"))
        ms5 = ms4.multiget(list("cgko"))
        ms6 = ms4.multiget(slice(2, 15, 4))
        filt = np.zeros((ms4.A.get_ncols(),), dtype=bool)
        filt[2:15:4] = True
        ms7 = ms4.multiget(filt)
        for ms in (ms5, ms6, ms7):
            self.assertEqual(list(ms.keys()), list("ABCD"))
            for dskey in "ABCD":
                self.assertEqual(list(getattr(ms, dskey).keys()), list("cgko"))
        ms8 = ms4.multiget("c")
        ms9 = ms4.multiget(["c"])
        for ms in (ms8, ms9):
            self.assertEqual(list(ms.keys()), list("ABCD"))
            for dskey in "ABCD":
                self.assertEqual(list(getattr(ms, dskey).keys()), list("c"))

    def test_copy(self):
        dd = {_k: self.get_basic_dataset(_i) for _i, _k in enumerate("bac")}
        ms = Multiset(dd)
        ms1 = ms.copy()
        self.assertEqual(list(ms1.keys()), ["b", "a", "c"])
        self.assertTrue((ms1.a == ms.a).all(axis=None))
        self.assertTrue((ms1.b == ms.b).all(axis=None))
        self.assertTrue((ms1.c == ms.c).all(axis=None))
        self.assertFalse(ms1.is_locked())
        ms._lock()
        ms2 = ms.copy()
        self.assertTrue(ms2.is_locked())

    def almost_eq(self, ar1, ar2, places=5):
        for v1, v2 in zip(ar1, ar2):
            self.assertAlmostEqual(v1, v2, places)

    def test_flatten(self):
        accuracy = 7
        message_types = [
            "NEW",
            "EXECUTE",
            "NEW",
            "EXECUTE",
            "EXECUTE",
            "EXECUTE",
            "EXECUTE",
            "CANCEL",
            "EXECUTE",
            "EXECUTE",
            "EXECUTE",
            "CANCEL",
        ]
        order_ids = [1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 1]
        milliseconds = [50, 70, 72, 75, 90, 88, 95, 97, 98, 115, 116, 120]
        shares = [0, 200, 0, 500, 100, 400, 100, 0, 300, 150, 150, 0]
        dat = Dataset(
            dict(
                message_type=message_types,
                order_id=order_ids,
                millisecond=milliseconds,
                shares=shares,
            )
        )
        dat = dat[["order_id", "message_type", "millisecond", "shares"]]
        gb = dat.groupby("order_id")
        ms1 = gb[["millisecond", "shares"]].aggregate(["sum", "mean"])
        # Flatten horizontally
        f1 = ms1.flatten()
        self.assertEqual(
            f1.keys(),
            [
                "order_id",
                "Sum_millisecond",
                "Sum_shares",
                "Mean_millisecond",
                "Mean_shares",
            ],
        )
        self.assertTrue((f1.Sum_millisecond == [410, 676]).all(axis=None))
        # Flatten vertically
        f2 = ms1.flatten(horizontal=False)
        self.assertEqual(f2.keys(), ["order_id", "Column", "millisecond", "shares"])
        self.almost_eq(f2.millisecond, [410.0, 676.0, 82.0, 96.57142857], accuracy)
        ms2 = gb[["millisecond", "shares"]].aggregate(["min", "max"])
        # Flatten multiset containing multisets
        ms3 = Multiset({"ms1": ms1, "ms2": ms2})
        f3 = ms3.flatten()
        self.assertEqual(
            f3.keys(),
            [
                "order_id",
                "ms1_Sum_millisecond",
                "ms1_Sum_shares",
                "ms1_Mean_millisecond",
                "ms1_Mean_shares",
                "ms2_Min_millisecond",
                "ms2_Min_shares",
                "ms2_Max_millisecond",
                "ms2_Max_shares",
            ],
        )
        self.assertTrue((f3.ms1_Sum_millisecond == [410, 676]).all(axis=None))
        f3v = ms3.flatten(horizontal=False)
        self.assertTrue(
            (
                f3v.Column
                == [
                    b"ms1_Sum",
                    b"ms1_Sum",
                    b"ms1_Mean",
                    b"ms1_Mean",
                    b"ms2_Min",
                    b"ms2_Min",
                    b"ms2_Max",
                    b"ms2_Max",
                ]
            ).all(axis=None)
        )
        # Flatten multiset containing multisets and a dataset
        ds = gb[["millisecond", "shares"]].std()
        ms4 = Multiset({"ms1": ms1, "ms2": ms2, "Std": ds})
        f4 = ms4.flatten()
        f4.label_remove()
        self.assertEqual(
            f4.keys(),
            [
                "order_id",
                "ms1_Sum_millisecond",
                "ms1_Sum_shares",
                "ms1_Mean_millisecond",
                "ms1_Mean_shares",
                "ms2_Min_millisecond",
                "ms2_Min_shares",
                "ms2_Max_millisecond",
                "ms2_Max_shares",
                "Std_millisecond",
                "Std_shares",
            ],
        )
        self.assertTrue((f4.ms1_Sum_millisecond == [410, 676]).all(axis=None))
        f4 = ms4.flatten(horizontal=False)
        self.assertAlmostEqual(f4[9, "millisecond"], 15.4903964, places=5)
        self.assertAlmostEqual(f4[9, "shares"], 148.4042099, places=5)
        # Flatten multiset containing multisets and a dataset with non-matching column
        ds1 = gb[["shares"]].std()
        ms5 = Multiset({"ms1": ms1, "ms2": ms2, "Std": ds1})
        f5 = ms5.flatten()
        f5.label_remove()
        self.assertEqual(
            f5.keys(),
            [
                "order_id",
                "ms1_Sum_millisecond",
                "ms1_Sum_shares",
                "ms1_Mean_millisecond",
                "ms1_Mean_shares",
                "ms2_Min_millisecond",
                "ms2_Min_shares",
                "ms2_Max_millisecond",
                "ms2_Max_shares",
                "Std_shares",
            ],
        )
        self.assertTrue((f5.ms1_Sum_millisecond == [410, 676]).all(axis=None))
        with self.assertRaises(ValueError):
            ms5.flatten(horizontal=False)


if __name__ == "__main__":
    tester = unittest.main()



import numpy as np
import pandas as pd
import riptable as rt
from typing import Tuple

import pytest
from numpy.testing import assert_array_almost_equal

import riptable as rt


def count(df, data=None):
    return df.count() if data is None else df.count()  ##this is different


def sum(df, data=None):
    return df.sum() if data is None else df.sum(data)


def mean(df, data=None):
    return df.mean() if data is None else df.mean(data)


def median(df, data=None):
    return df.median() if data is None else df.median(data)


def min(df, data=None):
    return df.min() if data is None else df.min(data)


def max(df, data=None):
    return df.max() if data is None else df.max(data)


def prod(df, data=None):
    return df.prod() if data is None else df.prod(data)


def var(df, data=None):
    return df.var() if data is None else df.var(data)


def quantile(df, data=None, q=None, **kwargs):
    return df.quantile(q=q, **kwargs) if data is None else df.quantile(data, q=q, **kwargs)


def cumsum(df, data=None):
    return df.cumsum() if data is None else df.cumsum(data)


def cumprod(df, data=None):
    return df.cumprod() if data is None else df.cumprod(data)


def cummax(df, data=None):
    return df.cummax() if data is None else df.cummax(data)


def cummin(df, data=None):
    return df.cummin() if data is None else df.cummin(data)


functions = [
    count,
    sum,
    mean,
    median,
    min,
    max,
    # prod,
    var,
    quantile,
    cumsum,
    cumprod,
    # cummax,
    # cummin
]

functions_str = [
    "count",
    "sum",
    "mean",
    "median",
    "min",
    "max",
    # 'prod',
    "var",
    "quantile",
    "cumsum",
    "cumprod",
    # 'cummax',
    # 'cummin'
]

# list of (rt_cat_kwargs, pd_gb_kwargs)
functions_kwargs = [
    ({}, {}),
    ({}, {}),
    ({}, {}),
    ({}, {}),
    ({}, {}),
    ({}, {}),
    # ({}, {}),
    ({}, {}),
    ({"q": 0.1743}, {"q": 0.1743, "interpolation": "midpoint"}),
    ({}, {}),
    ({}, {}),
    # ({}, {}),
    # ({}, {}),
]

type_list = [
    # np.bool,          ## not a numeric type
    np.intc,
    np.intp,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    # np.float16,       ## not supported
    np.float32,
    np.float64,
    # np.complex64,     ## not supported
    # np.complex128     ## not supported
]


class TestCategoricalSimpleGroupBy:
    def groupby_func(self, df, fn, kwargs, data):
        return fn(df, data=data, **kwargs)

    @pytest.mark.parametrize(
        ("func", "func_name", "func_kwargs"),
        [pytest.param(x, y, z, id=y) for (x, y, z) in zip(functions, functions_str, functions_kwargs)],
    )
    @pytest.mark.parametrize("data_dtype", type_list)
    def test_single_col_categoricals(self, func, func_name: str, func_kwargs: Tuple[dict, dict], data_dtype):
        values = [0, 1, 1, 2, 2, 2, 3, 3, 3, 4]
        bin_ids = ["a", "b", "c", "d", "e"]
        # data = np.random.rand(10) + np.random.randint(0, 10, size=10)
        data = np.array(
            [
                7.19200901,
                0.14907245,
                2.28258397,
                5.07872708,
                0.76125165,
                1.32797916,
                3.40280423,
                4.48942476,
                6.98713656,
                4.39541456,
            ]
        )

        data = rt.FastArray(data, dtype=data_dtype)

        map = {"vs": data, "ks": values}

        pd_data = pd.DataFrame(map).groupby(by="ks")
        rt_data = rt.Categorical(values=values, categories=bin_ids, base_index=0)

        cat_kwargs = func_kwargs[0]
        pd_kwargs = func_kwargs[1]

        pd_out = self.groupby_func(pd_data, func, pd_kwargs, None)
        rt_out = self.groupby_func(rt_data, func, cat_kwargs, data)

        col_index = "Count" if func_name == "count" else 0

        assert_array_almost_equal(rt_out[col_index], pd_out["vs"].values, decimal=3)



import unittest

import riptable as rt


class AccumTable_Test(unittest.TestCase):
    def test_accum_cols(self):

        num_rows = 10
        data = rt.Dataset(
            {
                "Symb": rt.Cat(["A", "B"] * int(num_rows / 2)),
                "Count": rt.full(num_rows, 1.0),
                "PlusMinus": [1.0, -1.0] * int(num_rows / 2),  # Added to handle edge case of zero footer
            }
        )

        accum = rt.accum_cols(data.Symb, [data.Count, data.PlusMinus], ["Count", "PlusMinus"])
        accum_expected = rt.Dataset({"Symb": ["A", "B"], "Count": [5.0, 5.0], "PlusMinus": [5.0, -5.0]})
        accum_expected.footer_set_values("Total", {"Symb": "Total", "Count": 10.0, "PlusMinus": 0.0})
        self.assertTrue((accum == accum_expected).all(axis=None))

    def test_accum_cols_multikey(self):
        num_rows = 12
        data = rt.Dataset(
            {
                "Symb": rt.Cat(["A", "B"] * int(num_rows / 2)),
                "Exch": rt.Cat(["X", "Y", "Y", "X"] * int(num_rows / 4)),
                "Count": rt.full(num_rows, 1.0),
                "PlusMinus": [1.0, -1.0] * int(num_rows / 2),
            }
        )
        data.MultiKeyCat = rt.Cat([data.Symb, data.Exch])

        accum = rt.accum_cols(data.MultiKeyCat, [data.Count, data.PlusMinus], ["Count", "PlusMinus"])
        accum_expected = rt.Dataset(
            {
                "Symb": ["A", "B", "A", "B"],
                "Exch": ["X", "Y", "Y", "X"],
                "Count": [3.0, 3.0, 3.0, 3.0],
                "PlusMinus": [3.0, -3.0, 3.0, -3.0],
            }
        )
        accum_expected.footer_set_values("Total", {"Exch": "Total", "Count": 12.0, "PlusMinus": 0.0})

        self.assertTrue((accum == accum_expected).all(axis=None))

    # When a raw FA is passed as the pointers instead of Categorical
    def test_accum_cols_noncat(self):
        num_rows = 10
        pointer = rt.FA([0, 1] * int(num_rows / 2))
        count = rt.full(num_rows, 1.0)

        accum = rt.accum_cols(pointer, count)
        accum_expected = rt.Dataset({"YLabel": [0, 1], "col0": [5.0, 5.0]})
        accum_expected.footer_set_values("Total", {"YLabel": "Total", "col0": 10.0})

        self.assertTrue((accum == accum_expected).all(axis=None))

    # Test basic accum_ratiop
    def test_accum_ratiop(self):

        num_rows = 12
        data = rt.Dataset(
            {
                "Symb": rt.Cat(["A", "A", "A", "B"] * int(num_rows / 4)),
                "Exch": rt.Cat(["Z", "Z", "X", "X"] * int(num_rows / 4)),
                "Count": rt.full(num_rows, 1.0),
            }
        )

        # Invalid input
        with self.assertRaises(ValueError, msg=f"Failed to raise an error when passing invalid norm_by arg"):
            rt.accum_ratiop(data.Symb, data.Exch, data.Count, norm_by="z")

        # Ratio within total
        accum = rt.accum_ratiop(data.Symb, data.Exch, data.Count, norm_by="T")
        accum_expected = rt.Dataset(
            {
                "Symb": ["A", "B"],
                "X": [25.0, 25.0],
                "Z": [50.0, 0.0],
                "TotalRatio": [75.0, 25.0],
                "Total": [9.0, 3.0],
            }
        )
        accum_expected.footer_set_values(
            "TotalRatio",
            {"Symb": "TotalRatio", "X": 50.0, "Z": 50.0, "TotalRatio": 100.0},
        )
        accum_expected.footer_set_values("Total", {"Symb": "Total", "X": 6.0, "Z": 6.0, "Total": 12.0})
        self.assertTrue((accum == accum_expected).all(axis=None))

        # Ratio within columns
        accum = rt.accum_ratiop(data.Symb, data.Exch, data.Count, norm_by="c")
        accum_expected = rt.Dataset(
            {
                "Symb": ["A", "B"],
                "X": [50.0, 50.0],
                "Z": [100.0, 0.0],
                "TotalRatio": [75.0, 25.0],
                "Total": [9.0, 3.0],
            }
        )
        accum_expected.footer_set_values(
            "TotalRatio",
            {"Symb": "TotalRatio", "X": 100.0, "Z": 100.0, "TotalRatio": 100.0},
        )
        accum_expected.footer_set_values("Total", {"Symb": "Total", "X": 6.0, "Z": 6.0, "Total": 12.0})
        self.assertTrue((accum == accum_expected).all(axis=None))


if __name__ == "__main__":
    tester = unittest.main()



import unittest

import pytest
from numpy import dtype
from numpy.testing import assert_array_equal

from riptable import *
from riptable.rt_enum import INVALID_DICT, TypeRegister
from riptable.rt_numpy import arange, ismember, assoc_copy


# TODO: Use pytest.mark.parametrize() here to iterate over the different dtypes
#       rather than looping over them within the test function.
def test_ismember_nums():
    allowed_int = "bhilqpBHILQP"
    allowed_float = "fd"
    allowed_types = allowed_int + allowed_float
    # allowed_types = 'lqfd'

    a_nums_list = [1, 2, 3, 4, 5]
    a_nums_tup = tuple(a_nums_list)

    b_nums_list = [1, 2]
    b_nums_tup = tuple(b_nums_list)

    a_nums = (
        [a_nums_list, a_nums_tup]
        + [np.array(a_nums_list, dtype=np.dtype(i)) for i in allowed_types]
        + [FA(a_nums_list, dtype=np.dtype(i)) for i in allowed_types]
    )
    b_nums = (
        [b_nums_list, b_nums_tup]
        + [np.array(b_nums_list, dtype=np.dtype(i)) for i in allowed_types]
        + [FA(b_nums_list, dtype=np.dtype(i)) for i in allowed_types]
    )

    correct_bool = [True, True, False, False, False]
    correct_idx = [0, 1]

    for a in a_nums:
        adt = None
        if hasattr(a, "dtype"):
            adt = a.dtype
        else:
            adt = type(a)
        for b in b_nums:
            bdt = None
            if hasattr(b, "dtype"):
                bdt = b.dtype
            else:
                bdt = type(b)

            bool_arr, idx_arr = ismember(a, b)

            assert_array_equal(
                bool_arr,
                correct_bool,
                err_msg=f"Boolean return was incorrect for ismember with a type {adt} b type {bdt}",
            )
            assert_array_equal(
                idx_arr[:2],
                correct_idx,
                err_msg=f"Index was incorrect for ismember with a type {adt} b type {bdt}",
            )

            # TODO: Use rt.isnan here to compare idx_arr[2:] (or just idx_arr) to an expected array; then we can use assert_array_equal()
            for idx in idx_arr[2:]:
                assert (
                    idx == INVALID_DICT[idx_arr.dtype.num]
                ), f"Index element was not invalid at index {idx} for ismember with a type {adt} b type {bdt}"


def test_ismember_strings():
    correct_bool = [True, True, False, False, False]
    correct_idx = [0, 1]

    bytes_list_a = ["a", "b", "c", "d", "e"]
    a_strings = [
        bytes_list_a,
        tuple(bytes_list_a),
        np.array(bytes_list_a),
        FA(bytes_list_a),
    ]
    # np.chararray(bytes_list_a) ]

    bytes_list_b = ["a", "b"]
    b_strings = [
        bytes_list_b,
        tuple(bytes_list_b),
        np.array(bytes_list_b),
        FA(bytes_list_b),
    ]
    # np.chararray(bytes_list_b) ]

    true_unicode_a = np.array(["a\u2082", "b\u2082", "c\u2082", "d\u2082", "e\u2082"])
    true_unicode_b = np.array(["a\u2082", "b\u2082"])

    # valid inputs, conversion happens when necessary
    for a in a_strings:
        adt = None
        if hasattr(a, "dtype"):
            adt = a.dtype
        else:
            adt = type(a)
        for b in b_strings:
            bdt = None
            if hasattr(b, "dtype"):
                bdt = b.dtype
            else:
                bdt = type(b)

            bool_arr, idx_arr = ismember(a, b)

            assert_array_equal(
                bool_arr,
                correct_bool,
                err_msg=f"Boolean return was incorrect for ismember with a type {adt} b type {bdt}",
            )
            assert_array_equal(
                idx_arr[:2],
                correct_idx,
                err_msg=f"Index was incorrect for ismember with a type {adt} b type {bdt}",
            )

            # TODO: Use rt.isnan here to compare idx_arr[2:] (or just idx_arr) to an expected array; then we can use assert_array_equal()
            for idx in idx_arr[2:]:
                assert (
                    idx == INVALID_DICT[idx_arr.dtype.num]
                ), f"Index element was not invalid at index {idx} for ismember with a type {adt} b type {bdt}"

    ## invalid first input
    # a = true_unicode_a
    # adt = "REAL unicode"
    # for b in b_strings:
    #    bdt = None
    #    if hasattr(b,'dtype'):
    #        bdt=b.dtype
    #    else:
    #        bdt = type(b)
    #    with pytest.raises(TypeError, msg=f"Failed to raise error between {adt} and {bdt} when unable to convert unicode."):
    #        bool_arr, idx_arr = ismember(a,b)

    ## invalid second input
    # b = true_unicode_b
    # bdt = "REAL unicode"
    # for a in a_strings:
    #    bdt = None
    #    if hasattr(a,'dtype'):
    #        adt=a.dtype
    #    else:
    #        adt = type(a)
    #    with pytest.raises(TypeError, msg=f"Failed to raise error between {adt} and {bdt} when unable to convert unicode."):
    #        bool_arr, idx_arr = ismember(a,b)

    ## both inputs invalid
    # a = true_unicode_a
    # b = true_unicode_b
    # with pytest.raises(TypeError, msg=f"Failed to raise error between {a.dtype} and {b.dtype} when unable to convert unicode."):
    #    bool_arr, idx_arr = ismember(a,b)


def test_ismember_categorical():
    for b_index_c in [0, 1]:
        for b_index_d in [0, 1]:

            # string values, both base indices
            c = TypeRegister.Categorical(
                np.random.choice(["a", "b", "c", "d", "e", "f"], 15),
                base_index=b_index_c,
            )
            d = TypeRegister.Categorical(np.random.choice(["a", "b", "c"], 10), base_index=b_index_d)
            cs, ds = c.as_string_array, d.as_string_array

            b, f = ismember(c, d)
            bs, fs = ismember(cs, ds)
            assert_array_equal(b, bs)
            assert_array_equal(int8(f), fs)

            b, f = ismember(d, c)
            bs, fs = ismember(ds, cs)
            assert_array_equal(b, bs)
            assert_array_equal(int8(f), fs)

            # codes, string values, both base indices

        c = TypeRegister.Categorical(np.random.choice(["a", "b", "c", "d", "e", "f"], 15), base_index=b_index_c)
        d = TypeRegister.Categorical(np.random.choice(["a", "b", "c"], 10), ["a", "b", "c"], base_index=1)
        cs, ds = c.as_string_array, d.as_string_array
        b, f = ismember(c, d)
        bs, fs = ismember(cs, ds)
        assert_array_equal(b, bs)
        assert_array_equal(int8(f), fs)

        b, f = ismember(d, c)
        bs, fs = ismember(ds, cs)
        assert_array_equal(b, bs)
        assert_array_equal(int8(f), fs)

    c = Categorical(np.random.choice(["a", "b", "c"], 15))
    with pytest.raises(TypeError):
        b, idx = ismember(c, arange(3))
    # TypeRegister.Categorical.TestIsMemberVerbose = True
    # c = TypeRegister.Categorical(np.random.choice(['a','b','c', 'd', 'e', 'f'],15), base_index=1)
    # _ = ismember(c, ['a','b'])
    # self.assertEqual("newcat", TypeRegister.Categorical._test_cat_ismember)


def test_ismember_dataset():
    ds = Dataset(dict(cat=Categorical([1, 2, 3] * 3, ["a", "b", "c"]), ints=np.arange(9)))
    indexer = [2, 4, 6]
    bools, key = ismember(ds[indexer, :], ds)
    assert bools.all()
    assert_array_equal(key, indexer)

    bools, key = ismember(ds[:3, :], ds[2:, :])
    assert key.tolist() == [key.inv, key.inv, 0]


def test_ismember_2d_array():
    arr = np.arange(10).reshape(5, 2)
    indexer = [2, 4]
    bools, key = ismember(arr[indexer, :], arr)
    assert bools.all()
    assert_array_equal(key, indexer)

    bools, key = ismember(arr[:3, :], arr[2:, :])
    assert key.tolist() == [key.inv, key.inv, 0]


class Ismember_Test(unittest.TestCase):
    # TODO: Convert this test to a module-level, pytest-style test.
    def test_ismember_categorical_numeric(self):
        c = Categorical([1, 2, 3, 1, 2, 3, 1, 2, 4])
        f = FastArray([1, 2, 3], dtype=np.int64)
        b, idx = ismember(c, f)
        self.assertTrue(bool(np.all(b[:-1])))
        self.assertFalse(b[-1], False)
        self.assertTrue(bool(np.all(idx[:-1] == tile(FA([0, 1, 2]), 3)[:-1])))
        self.assertTrue(idx.isnan()[-1])
        f = FastArray(["a", "b", "c"])
        with pytest.raises(TypeError):
            b, idx = ismember(c, f)

        c = Categorical([1.0, 2.0, 3.0, 1.0, 2.0, 3.0, 1.0, 2.0, 4.0])
        f = FastArray([1, 2, 3], dtype=np.float64)
        b, idx = ismember(c, f)
        self.assertTrue(bool(np.all(b[:-1])))
        self.assertFalse(b[-1], False)
        self.assertTrue(bool(np.all(idx[:-1] == tile(FA([0, 1, 2]), 3)[:-1])))
        self.assertTrue(idx.isnan()[-1])
        f = FastArray(["a", "b", "c"])
        with pytest.raises(TypeError):
            b, idx = ismember(c, f)

        c = Categorical([np.random.choice(["a", "b", "c"], 10), arange(10)])
        with pytest.raises(TypeError):
            b, idx = ismember(c, f)


@pytest.mark.skip("RIP-364: This test is empty and needs to be implemented.")
def test_ismember_categorical_codes():
    pass
    # c_codes = np.random.randint(0,1000,1000)
    # d_codes = np.random.randint(0,1000,500)

    # for_cats = arange(5000,6000)
    # cats = (for_cats.astype('S'), True)
    # unicats = (for_cats.astype('U'), False)

    # for c1 in [cats, unicats]:
    #    c = Categorical(c_codes, c1[0], convert_string_to_bytes=c1[1])
    #    cs = c.as_string_array

    #    for c2 in [cats, unicats]:
    #        d = Categorical(d_codes, c2[0], convert_string_to_bytes=c2[1])
    #        ds = d.as_string_array

    #        b, f = ismember(c, d)
    #        bs, fs = ismember(cs, ds)
    #        self.assertTrue(bool(np.all(b==bs)))
    #        self.assertTrue(bool(np.all(int8(f)==fs)))

    #        b, f = ismember(d, c)
    #        bs, fs = ismember(ds, cs)
    #        self.assertTrue(bool(np.all(b==bs)))
    #        self.assertTrue(bool(np.all(int8(f)==fs)))


def test_ismember_index_casting():
    casting_dict = {
        5: np.dtype(np.int8),
        29_000: np.dtype(np.int16),
        60_000: np.dtype(np.int32),
    }
    for sz, correct_type in casting_dict.items():
        a = np.full(sz, b"a").view(FA)
        b = FA(["a", "b"])
        _, idx = ismember(b, a)
        assert (
            idx.dtype == correct_type
        ), f"ismember with {sz} strings did not properly downcast index array to {correct_type}, got {idx.dtype} instead."


# TODO: RIP-364: This test doesn't have any assertions, so it's currently only checking whether the function excepts. Add assertions to check the ismember output for correctness."
def test_ismember_int_edges():
    # hit thresholds for a previous bug
    for a_size in [127, 129, 254, 256]:
        a = arange(a_size)
        for b_size in range(129):
            _, _ = ismember(a, arange(b_size))


# TODO: RIP-364: This test doesn't have any assertions, so it's currently only checking whether the function excepts. Add assertions to check the ismember output for correctness."
def test_ismember_diff_itemsize():
    a = FA([b"b", b"a", b"a", b"Inv", b"c", b"a", b"b"], dtype="S")
    b = FA([b"a", b"b", b"c"], dtype="S")
    bl, idx = ismember(a, b)
    correctbool = [True, True, True, False, True, True, True]
    correctidx = FA([1, 0, 0, int8.inv, 2, 0, 1], dtype=np.int8)
    boolmatch = bool(np.all(bl == [False, True, True]))
    idxmatch = bool(np.all(idx == [int8.inv, 1, 2]))


@pytest.mark.skip("RIP-364: This test is empty and needs to be implemented.")
def test_ismember_empty():
    """Test which checks how ismember handles empty input(s)."""
    pass


def test_ismember_multikey_errors():
    # types didn't match
    a_keys = [FA([1, 2, 3]), FA([1, 2, 3], dtype=np.float32)]
    b_keys = [FA([1, 2, 3]), FA(["a", "b", "c"])]
    with pytest.raises(TypeError):
        _, _ = ismember(a_keys, b_keys)

    # different number of key columns
    a_keys = [FA([1, 2, 3])]
    b_keys = [FA([1, 2, 3]), FA([1, 2, 3], dtype=np.float64)]
    with pytest.raises(ValueError):
        _, _ = ismember(a_keys, b_keys)


def test_ismember_multikey_single_key():
    # send single columns through regular ismember
    a_keys = [FA([1, 2, 3])]
    b_keys = [FA([1, 2])]
    c, d = ismember(a_keys, b_keys)
    correct_bool = [True, True, False]
    correct_idx = FastArray([0, 1, INVALID_DICT[d.dtype.num]], dtype=d.dtype)

    assert_array_equal(c, correct_bool, err_msg=f"Incorrect boolean array for single-key ismembermk")
    assert_array_equal(d, correct_idx, err_msg=f"Incorrect index array for single-key ismembermk")


def test_ismember_multikey_non_ndarray():
    # lists
    correct_bool = [True, True, False]
    correct_idx = FastArray([0, 1, INVALID_DICT[np.dtype(np.int8).num]], dtype=np.int8)

    a_keys = [[1, 2, 3], ["a", "b", "c"]]
    b_keys = [[1, 2], ["a", "b"]]
    c, d = ismember(a_keys, b_keys)
    assert_array_equal(c, correct_bool, err_msg=f"Incorrect boolean array for keys as lists ismembermk")
    assert_array_equal(d, correct_idx, err_msg=f"Incorrect index array for keys as lists ismembermk")

    # tuples
    a_keys = [tuple(key) for key in a_keys]
    b_keys = [tuple(key) for key in b_keys]
    c, d = ismember(a_keys, b_keys)
    assert_array_equal(
        c,
        correct_bool,
        err_msg=f"Incorrect boolean array for keys as tuples ismembermk",
    )
    assert_array_equal(d, correct_idx, err_msg=f"Incorrect index array for keys as tuples ismembermk")

    # nparray
    a_keys = [np.array(key) for key in a_keys]
    b_keys = [np.array(key) for key in b_keys]
    c, d = ismember(a_keys, b_keys)
    assert_array_equal(
        c,
        correct_bool,
        err_msg=f"Incorrect boolean array for keys as np arrays ismembermk",
    )
    assert_array_equal(
        d,
        correct_idx,
        err_msg=f"Incorrect index array for keys as np arrays ismembermk",
    )


# TODO: RIP-364: This test doesn't have any assertions, so it's currently only checking whether the function excepts. Add assertions to check the ismember output for correctness."
def test_ismember_multikey_unicode():
    a_keys = [
        np.array(["a\u2082", "b\u2082", "c\u2082", "d\u2082", "e\u2082"]),
        [1, 2, 3, 4, 5],
    ]
    b_keys = [np.array(["a\u2082", "b\u2082"]), [1, 2]]
    b, idx = ismember(a_keys, b_keys)
    # with pytest.raises(TypeError):
    #    _, _ = ismember(a_keys, b_keys)


def test_ismember_align_multikey():
    correct_bool = FastArray([True, True, True, False, False])
    correct_idx = FastArray([0, 1, 2, int8.inv, int8.inv], dtype=np.int8)

    # bytes / unicode both upcast
    a_keys = [arange(5), FastArray([b"a", b"b", b"c", b"d", b"e"], dtype="S5")]
    b_keys = [arange(3), FastArray(["a", "b", "c"], dtype="U4", unicode=True)]
    b, idx = ismember(a_keys, b_keys)
    assert_array_equal(b, correct_bool)
    # NOTE: flip to numpy because FastArray is sentinel-aware
    assert_array_equal(idx._np, correct_idx._np)
    assert a_keys[1].dtype.char == "S"

    # bytes / Categorical unicode
    a_keys = [arange(5), FastArray(["a", "b", "c", "d", "e"], dtype="S5")]
    b_keys = [
        arange(3),
        Categorical(FastArray(["a", "b", "c"], dtype="U4", unicode=True), unicode=True),
    ]
    b, idx = ismember(a_keys, b_keys)
    assert_array_equal(b, correct_bool)
    # NOTE: flip to numpy because FastArray is sentinel-aware
    assert_array_equal(idx._np, correct_idx._np)

    # unicode / Categorical
    a_keys = [arange(5), FastArray(["a", "b", "c", "d", "e"], dtype="U5", unicode=True)]
    b_keys = [
        arange(3),
        Categorical(FastArray(["a", "b", "c"], dtype="U4", unicode=True), unicode=True),
    ]
    b, idx = ismember(a_keys, b_keys)
    assert_array_equal(b, correct_bool)
    # NOTE: flip to numpy because FastArray is sentinel-aware
    assert_array_equal(idx._np, correct_idx._np)

    # different numeric types
    a_keys = [
        arange(5, dtype=np.float64),
        FastArray(["a", "b", "c", "d", "e"], dtype="U5", unicode=True),
    ]
    b_keys = [
        arange(3),
        Categorical(FastArray(["a", "b", "c"], dtype="U4", unicode=True), unicode=True),
    ]
    b, idx = ismember(a_keys, b_keys)
    assert_array_equal(b, correct_bool)
    # NOTE: flip to numpy because FastArray is sentinel-aware
    assert_array_equal(idx._np, correct_idx._np)

    # string / non-string
    a_keys = [
        arange(5).astype("S"),
        FastArray(["a", "b", "c", "d", "e"], dtype="U5", unicode=True),
    ]
    b_keys = [
        arange(3),
        Categorical(FastArray(["a", "b", "c"], dtype="U4", unicode=True), unicode=True),
    ]
    with pytest.raises(TypeError):
        b, idx = ismember(a_keys, b_keys)

    # multikey categorical, no expand array
    a_keys = [
        arange(5).astype("S"),
        FastArray(["a", "b", "c", "d", "e"], dtype="U5", unicode=True),
    ]
    b_keys = [
        arange(3),
        Categorical(
            [FastArray(["a", "b", "c"], dtype="U4", unicode=True), arange(3)],
            unicode=True,
        ),
    ]
    with pytest.raises(TypeError):
        b, idx = ismember(a_keys, b_keys)
    with pytest.raises(TypeError):
        b, idx = ismember(b_keys, a_keys)

    # unsupported object array
    a_keys = [arange(5).astype("O"), FastArray(["a", "b", "c", "d", "e"], dtype="S5")]
    b_keys = [arange(3), FastArray(["a", "b", "c"], dtype="U4", unicode=True)]

    with pytest.raises(TypeError):
        b, idx = ismember(a_keys, b_keys)


def test_ismember_categorical_with_invalid():
    c1 = Categorical(["a", "b", "c"])
    c2 = Categorical(["a", "b", "c", "d", "c"])
    c1[0] = 0
    c2[3] = 0
    # This is a little hacky because it relies on the implementation detail
    np.asarray(c1)[1] = c1.inv
    np.asarray(c2)[2] = c2.inv
    is_in, idx = ismember(c1, c2)
    expected_is_in = FastArray([False, False, True])
    assert_array_equal(is_in, expected_is_in)
    expected_idx = FastArray([0, 0, 4], dtype=idx.dtype)
    expected_idx[0:2] = expected_idx.inv
    assert_array_equal(idx, expected_idx)


@pytest.mark.parametrize(
    "a, b, isin, indexer",
    [
        (FA([0, 1, 2]), FA([]), FastArray([False] * 3), tile(INVALID_DICT[np.dtype("int32").num], 3).astype("int32")),
        (FA([]), FA([0, 1, 2]), FA([], dtype=bool), FA([], dtype="int32")),
    ],
)
def test_ismember_type_for_empties(a, b, isin, indexer):
    result = ismember(a, b)
    assert_array_equal(isin, result[0])
    assert_array_equal(indexer, result[1])
    assert result[1].isna().all()


@pytest.mark.parametrize(
    "dt1, dt2, expected_failure",
    [
        ("int8", "float32", False),
        ("int8", "int8", False),
        ("int16", "int64", False),
        ("int32", "float32", False),
        ("int32", "float64", False),
        ("int32", "int64", False),
        ("int64", "float32", False),
        ("int64", "float64", False),
        ("uint8", "float32", False),
        ("uint8", "int8", False),
        ("uint64", "float32", False),
        ("uint64", "float64", False),
        ("uint64", "int64", False),
        ("uint64", "str", True),
    ],
)
def test_ismember_lossless_check(dt1, dt2, expected_failure):
    def default_value(dt):
        return 0 if np.issubdtype(dt, np.number) else ""

    fa1 = FA([default_value(dt1)], dtype=dt1)
    fa2 = FA([default_value(dt2)], dtype=dt2)

    actual_failure = False
    try:
        ismember(fa1, fa2)
    except TypeError:
        actual_failure = True
    assert actual_failure == expected_failure

    actual_failure = False
    try:
        ismember(fa2, fa1)
    except TypeError:
        actual_failure = True
    assert actual_failure == expected_failure


def test_is_member_casting_uints():
    # RIP-254:  As floats n and n+1 are the same. We check that we don't identify n and n+1 in ismember
    n = 2**53
    a = FastArray([n + 2, n + 1, n, n - 1], dtype="uint64")
    b = FastArray([n])
    res = ismember(a, b)[0]
    assert_array_equal(res, [False, False, True, False])
    res = ismember(b, a)[1]
    assert (res == 2).all()


def test_ismember_empty_key():
    a = []
    b = [1, 2, 3]
    ismem, ix = ismember(a, b)
    assert_array_equal(ismem, FA([], dtype=bool))
    assert_array_equal(ix, FA([], dtype="int32"))


def test_ismember_empty_multi_key():
    a = [[], []]
    b = [[1, 2, 3], [4, 5, 6]]
    ismem, ix = ismember(a, b)
    assert_array_equal(ismem, FA([], dtype=bool))
    assert_array_equal(ix, FA([], dtype="int32"))


def test_is_member_casting_big_uints():
    # RIP-254
    n = 2**63
    a = FastArray([n - 1], dtype="int64")
    b = FastArray([3, n - 1], dtype="uint64")
    res = ismember(a, b)[0]
    #    assert_array_equal(res, [False, True])
    assert_array_equal(res, [True])
    res = ismember(b, a)[1]
    assert_array_equal(res, FastArray([-128, 0], dtype="int8"))
    res = ismember(a, b)[1]
    assert (res == 1).all()


@pytest.mark.parametrize(
    "values, expected",
    [
        ([1, 2, 3], [INVALID_DICT[9], 2, INVALID_DICT[9]]),
        ([1.0, 2.0, 3.0], [np.nan, 2, np.nan]),
        (np.array([1.0, 2.0, 3.0]), [np.nan, 2, np.nan]),
    ],
)
def test_assoc_copy_array_input(values, expected):
    result = assoc_copy([0, 2, 4], [1, 2, 3], values)
    assert_array_equal(result, FastArray(expected))


def test_assoc_copy_dataset_input():
    values = Dataset({"a": [1, 2, 3, 4], "b": [0.1, 0.2, 0.3, 0.4]})
    result = assoc_copy([0, 2, 4], [1, 2, 3, 4], values)
    expected = Dataset({"a": [INVALID_DICT[9], 2, 4], "b": [np.nan, 0.2, 0.4]})
    assert list(result) == list(expected)
    for k in result:
        assert_array_equal(result[k], FastArray(expected[k]))


if __name__ == "__main__":
    tester = unittest.main()



"""
Helper functions for generating random data for use in implementing unit tests for riptable.
"""
__all__ = ["create_test_dataset"]

import datetime as dt

import numpy as np
from numpy.random import Generator

from .. import Categorical, Dataset, Date, DateTimeNano, TimeSpan


def create_test_dataset(
    *,
    rng: Generator,
    rowcount: int,
    include_dict_cat: bool = False,
    include_date: bool = False,
    include_datetime_nano: bool = True,
    include_timespan: bool = True,
) -> Dataset:
    test_ds = Dataset()

    # Add deliverable_symbol, drawn as a subset of a few (say, 3-5) symbols from a larger set of symbols.
    # Tile it, truncate to the desired rowcount, then turn it into a Categorical
    # TODO: Add more symbols to this list for better testing; the symbols themselves don't matter,
    #       it's more of a set theory thing -- we want to get coverage around cases where there's
    #       (no, partial, complete) overlap in symbols between sections, etc.
    deliv_symbols = np.array(["CBO", "CBX", "ZBZX", "ZJZZT", "ZTEST", "ZVV", "ZVZZT", "ZWZZT", "ZXZZT"])
    symbol_count = 3
    # Choose N unique symbols
    attempts = 100
    symbol_idxs = None
    for _ in range(attempts):
        # Generate N random indices in the range of the symbol set.
        rand_idxs = rng.integers(0, len(deliv_symbols), size=symbol_count)

        # If the generated indices are all unique, save this array and break out of the loop.
        if len(np.unique(rand_idxs)) == symbol_count:
            symbol_idxs = rand_idxs
            break

    # If we couldn't get a unique set of indices, we can't proceed.
    if symbol_idxs is None:
        raise Exception("Unable to draw a unique set of indices.")

    # Draw a fancy index (that we'll then use to fancy-index our symbol indices) from a distribution that makes
    # some elements more likely to appear than others.
    bucket_weights = np.arange(symbol_count + 1, 1, -1)
    # TODO: If we generalize this to a function, add a check here for negative bucket weights
    total_bucket_weight = np.sum(bucket_weights)
    cum_bucket_weights = bucket_weights.cumsum(dtype=np.min_scalar_type(total_bucket_weight))
    # 'endpoint=False' here because we specify 'side=right' for searchsorted below.
    # TODO: This won't be correct if we have any zero-weighted buckets, it could mean we draw from a zero-weighted bucket
    #       instead of an adjacent non-zero-weighted bucket.
    samples = rng.integers(0, total_bucket_weight, endpoint=False, size=rowcount)
    sampled_buckets = np.searchsorted(cum_bucket_weights, samples, side="right")

    # Get the set of unique symbols then create the Categorical column.
    # N.B. If desired, it'd be a fairly simple change to randomly include more of the symbols
    #      so they'll be included in the Categorical but not actually be used -- might be useful
    #      for testing loading/hstacking.
    test_ds.deliverable_symbol = Categorical(sampled_buckets, categories=deliv_symbols[symbol_idxs])

    # Create a dictionary-mode Categorical (from ISO3166 data).
    if include_dict_cat:
        category_dict = {
            "IRL": 372,
            "USA": 840,
            "AUS": 36,
            "HKG": 344,
            "JPN": 392,
            "MEX": 484,
            "KHM": 116,
            "THA": 764,
            "JAM": 388,
            "ARM": 51,
        }
        # The values for the Categorical's backing array.
        # This includes some value(s) not in the dictionary and not all values in the dictionary are used here.
        cat_values = [36, 36, 344, 840, 840, 372, 840, 372, 840, 124, 840, 124, 36, 484]
        # calculate cat length
        tile_num = np.math.ceil(rowcount / len(cat_values))
        cat_values = np.tile(cat_values, tile_num)[:rowcount]
        test_ds.country = Categorical(cat_values, categories=category_dict)

    # Add dates and time spans from rowcount days ago
    offset_day_range = 1_095  # offset day range by three years
    offset_days = rng.integers(-offset_day_range, offset_day_range)
    start_date = dt.date.today() + dt.timedelta(days=int(offset_days - rowcount))
    end_date = start_date + dt.timedelta(days=rowcount)

    np_date = np.arange(start_date, end_date, dt.timedelta(days=1)).astype("<M8[D]")
    dtns = DateTimeNano(np_date, from_tz="GMT")  # TODO: exercise other timezones

    if include_datetime_nano:
        test_ds.datetime_nano = dtns
        # TODO enable DTN categoricals after resolving rt.Cat inconsistencies
        # test_ds.datetime_nano_cat = dtns_cat = rt.Categorical(test_ds.datetime_nano)
        # dtns_cat[0] = dtns_cat[1] = dtns_cat[rowcount-1] = 0  # set invalids

    if include_date:
        test_ds.np_date = np_date
        test_ds.date = Date(test_ds.np_date)
        test_ds.date_cat = Categorical(test_ds.date)

    if include_timespan:
        test_ds.time_span = TimeSpan([dtns[i - 1] - dtns[1] for i, _ in enumerate(dtns)])
        # TODO enable TS categoricals after resolving rt.Cat inconsistencies
        # test_ds.time_span_cat = ts_cat = rt.Categorical(test_ds.time_span)
        # ts_cat[0] = ts_cat[1] = ts_cat[rowcount-1] = 0  # set invalids

    # Add opening price, drawn from non-negative floats.
    test_ds.open_price = rng.uniform(10.0, 1000.0, size=rowcount)

    # Add closing price; draw from lognormal distribution then multiply with opening price (and max(..., 0)).
    test_ds.close_price = np.maximum(0.0, test_ds.open_price * rng.lognormal(sigma=0.04, size=rowcount))

    # Add a volume column as uint32
    test_ds.volume = rng.integers(0, 10_000_000, size=rowcount, dtype=np.uint32)

    return test_ds
